searchNodes=[{"doc":"Used to generate documentation from your zig code and incorporate it into your Elixir documentation.Generally speaking, zig structs will be documented as Elixir modules and zig functions will be incorporated as Elixir functions.If you want to make the zig docgen process override the Elixir mix docs task, add this to your mix.exs:def project do [ ... aliases: [docs: &quot;zig_doc&quot;], ... ] end","ref":"Mix.Tasks.ZigDoc.html","title":"Mix.Tasks.ZigDoc","type":"task"},{"doc":"Inline NIF support for ZigMotivationZig is a general-purpose programming language designed for robustness, optimality, and maintainability.The programming philosophy of Zig matches up nicely with the programming philosophy of the BEAM VM and in particular its emphasis on simplicity and structure should very appealing to the practitioners of Elixir.The following features make Zig extremely amenable to inline language support in a BEAM language:simplicity. Zig's syntax is definable in a simple YACC document and Zig takes a stance against making its featureset more complex (though it may evolve somewhat en route to 1.0)Composability. Zig is unopinionated about how to go about memory allocations. Its allocator interface is very easily able to be backed by the BEAM's, which means that you have access to generic memory allocation strategies through its composable allocator scheme.C integration. It's very easy to design C-interop between Zig and C. In fact, Zig is likely to be an easier glue language for C ABIs than C.Basic usageIn the BEAM, you can define a NIF by consulting the following document and implementing the appropriate shared object/DLL callbacks. However, Zigler will take care of all of this for you.Simply use Zig in your module, providing the app atom in the property list.Then, use the sigil_Z/2 macro and write zig code. Any nifs you define should be preceded with the /// nif: function_name/arity zig docstring.Exampledefmodule MyModule do use Zig ~Z&quot;&quot;&quot; /// nif: my_func/1 fn my_func(val: i64) i64 { return val + 1; } &quot;&quot;&quot; endZig will automatically fill out the appropriate NIF C template, compile the shared object, and bind it into the module pre-compilation. In the case of the example, there will be a MyModule.my_func/1 function call found in the module.Zig will also make sure that your statically-typed Zig data are guarded when you marshal it from the dynamically-typed BEAM world. However, you may only pass in and return certain types. As an escape hatch, you may use the beam.term type which is equivalent to the ERLNIFTERM type. See erl_nif.Nerves SupportNerves is supported out of the box, and the system should cross-compile to arm ABI as necessary depening on what your nerves :target is. You may also directly specify a zig target using the use Zig, target: &lt;target&gt; option.EnvironmentSometimes, you will need to pass the BEAM environment (which is the code execution context, including process info, etc.) into the NIF function. In this case, you should pass it as the first argument, as a beam.env type value.Exampledefmodule MyModule do use Zig ~Z&quot;&quot;&quot; /// nif: my_func_with_env/1 fn my_func_with_env(env: beam.env, pid: beam.pid) void { var sendable_term: []u64 = &quot;ping&quot;[0..]; var msg = beam.make_slice(env, sendable_term); var res = e.enif_send(env, pid, env, msg); } &quot;&quot;&quot; endBring your own version of ZigIf you would like to use your system's local zig command, set the local_zig option in your use Zig statement.defmodule UsesLocalZig do use Zig, local_zig: true ~Z&quot;&quot;&quot; ...External LibrariesIf you need to bind static (*.a) or dynamic (*.so) libraries into your module, you may link them with the :libs argument.Note that zig statically binds shared libraries into the assets it creates. This simplifies deployment for you.Exampledefmodule Blas do use Zig, libs: [&quot;/usr/lib/x86_64-linux-gnu/blas/libblas.so&quot;], include: [&quot;/usr/include/x86_64-linux-gnu&quot;] ~Z&quot;&quot;&quot; const blas = @cImport({ @cInclude(&quot;cblas.h&quot;); ...Compilation assistanceIf something should go wrong, Zigler will translate the Zig compiler error into an Elixir compiler error, and let you know exactly which line in the ~Z block it came from.Syntactic SugarSome of the erlang nif terms can get unwieldy, especially in Zig, which prefers terseness. Each of the basic BEAM types is shadowed by a Zig type in the beam module. The beam struct is always imported into the header of the zig file used, so all zig code in the same directory as the module should have access to the beam struct if they @import(&quot;beam.zig&quot;)Importing filesIf you need to write code outside of the basic module (you will, for anything non-trivial), just place it in the same directory as your module.Example~Z&quot;&quot;&quot; const extra_code = @import(&quot;extra_code.zig&quot;); /// nif: use_extra_code/1 fn use_extra_code(val: i64) i64 { return extra_code.extra_fn(val); } &quot;&quot;&quot;If you would like to include a custom c header file, create an include/ directory inside your path tree and it will be available to zig as a default search path as follows:~Z&quot;&quot;&quot; const c = @cImport({ @cInclude(&quot;my_c_header.h&quot;); }); // nif: my_nif/1 ... &quot;&quot;&quot;If the c header defines extern functions, it's your responsibility to make sure those externed functions are available by compiling other c files or using a shared library.DocumentationUse the builtin zig /// docstring to write your documentation. If it's in front of the nif declaration, it will wind up in the correct place in your elixir documentation.See Zig.Doc for more information on how to document in zig and what to document. See Mix.Tasks.ZigDoc for information on how to get your Elixir project to incorporate zig documentation.TestsUse the builtin zig test keyword to write your internal zig unit tests. These can be imported into an ExUnit module by following this example:defmodule MyTest do use ExUnit.Case use Zig.Unit zigtest ModuleWithZigCode endSee Zig.Unit for more information.","ref":"Zig.html","title":"Zig","type":"module"},{"doc":"declares a string block to be included in the module's .zig source file. At least one of these blocks must define a nif.","ref":"Zig.html#sigil_Z/2","title":"Zig.sigil_Z/2","type":"macro"},{"doc":"like sigil_Z/2, but lets you interpolate values from the outside elixir context using string interpolation (the \#{value} form)","ref":"Zig.html#sigil_z/2","title":"Zig.sigil_z/2","type":"macro"},{"doc":"Code for interfacing with std.build.Builder, the interface for programattically invoking build code with the zig build command.","ref":"Zig.Builder.html","title":"Zig.Builder","type":"module"},{"doc":"","ref":"Zig.Builder.html#build/2","title":"Zig.Builder.build/2","type":"function"},{"doc":"","ref":"Zig.Builder.html#target_struct/2","title":"Zig.Builder.target_struct/2","type":"function"},{"doc":"","ref":"Zig.Builder.html#target_struct_from_env/1","title":"Zig.Builder.target_struct_from_env/1","type":"function"},{"doc":"Translates the docstrings from your Zig code into Elixir documentation.For instructions on how to incorporate this into an Elixir project, consult Mix.Tasks.ZigDocDocumentation formsCurrently, Zigler recognizes four types of code segments which should be documented.functionstypesvalueserrorsFunctionsfunctions have the following signature:pub fn &lt;identifier&gt;(&lt;arguments&gt;) &lt;type&gt; {and may have the property of being comptime which is due to either the function being itself a comptime function or it having a comptime argument.NB only pub functions are documented in Zigler, following the Elixir philosophy that only public functions should be documented.Typestypes have the following signature:pub const &lt;identifier&gt;=&lt;value&gt;;Valuesvalues can have one of the following signatures:pub const &lt;identifier&gt;=&lt;value&gt;;pub var &lt;identifier&gt;=&lt;value&gt;;Note that the constant value form is indistinguishable from the type form without doing a full parse and evaluation of the Zig code. In order to avoid doing this, to disambiguate between the two, you must prepend constant value docststrings with a !value token.Example/// !value /// /// a constant representing the value 47. pub const fortyseven = 47;Errorserrors appear inside special error struct and should be documented at the per-value level:Example/// docstring for this error struct (if desired) pub const my_error = error { /// docstring for ErrorEnum1 ErrorEnum1, /// docstring for ErrorEnum2 ErrorEnum2 }ScopeThis module will generate documentation from all Zig code that resides in the same code directory as the base module (or overridden directory, if applicable). Zig code in subdirectories will not be subjected to document generation.","ref":"Zig.Doc.html","title":"Zig.Doc","type":"module"},{"doc":"Adapter code for synchronous nifs","ref":"Zig.Nif.Synchronous.html","title":"Zig.Nif.Synchronous","type":"module"},{"doc":"Adapter code for test nifs.","ref":"Zig.Nif.Test.html","title":"Zig.Nif.Test","type":"module"},{"doc":"Generates code for threaded nifs.threaded functions require several parts to get right.a resource that holds space for arguments, result, and flags. This is held on to by both the calling functiona packer function which takes the beam arguments and shoves them into resource struct, then launches the function, returning the resource.a launcher function which runs the packer wrapping the errors from the launch function. The launch function must be a nif function, as it will be called from the BEAM.a harness function which is passed the resource struct, and is responsible for unwrapping beam terms into function parameters. This is what runs the nif function.a catch function which releases the resource reference, and signals to the parent process that it's finished.","ref":"Zig.Nif.Threaded.html","title":"Zig.Nif.Threaded","type":"module"},{"doc":"","ref":"Zig.Nif.Threaded.html#cache/1","title":"Zig.Nif.Threaded.cache/1","type":"function"},{"doc":"","ref":"Zig.Nif.Threaded.html#cache_cleanup/1","title":"Zig.Nif.Threaded.cache_cleanup/1","type":"function"},{"doc":"","ref":"Zig.Nif.Threaded.html#cache_ptr/1","title":"Zig.Nif.Threaded.cache_ptr/1","type":"function"},{"doc":"","ref":"Zig.Nif.Threaded.html#cache_struct/1","title":"Zig.Nif.Threaded.cache_struct/1","type":"function"},{"doc":"","ref":"Zig.Nif.Threaded.html#cleanup/1","title":"Zig.Nif.Threaded.cleanup/1","type":"function"},{"doc":"","ref":"Zig.Nif.Threaded.html#cleanup_fn/1","title":"Zig.Nif.Threaded.cleanup_fn/1","type":"function"},{"doc":"","ref":"Zig.Nif.Threaded.html#harness/1","title":"Zig.Nif.Threaded.harness/1","type":"function"},{"doc":"","ref":"Zig.Nif.Threaded.html#harness_fn/1","title":"Zig.Nif.Threaded.harness_fn/1","type":"function"},{"doc":"","ref":"Zig.Nif.Threaded.html#launcher/1","title":"Zig.Nif.Threaded.launcher/1","type":"function"},{"doc":"","ref":"Zig.Nif.Threaded.html#launcher_fn/1","title":"Zig.Nif.Threaded.launcher_fn/1","type":"function"},{"doc":"","ref":"Zig.Nif.Threaded.html#name/1","title":"Zig.Nif.Threaded.name/1","type":"function"},{"doc":"","ref":"Zig.Nif.Threaded.html#packer/1","title":"Zig.Nif.Threaded.packer/1","type":"function"},{"doc":"","ref":"Zig.Nif.Threaded.html#packer_fn/1","title":"Zig.Nif.Threaded.packer_fn/1","type":"function"},{"doc":"Generates code for yielding nifs.yielding functions require several parts to get right.a zig struct that holds the frame of the async function and also information to be filled in for the yielding allocator.a launcher function which serves as the interface between the C abi, which cannot tolerate async functions and zig function that is going to be called.a supervisor function which is responsible for resuming the async function. This is the event loop, but it uses a tail-call sort of a system that Erlang is used to.a harness function which takes care of informing the supervisor that target function has been completed.","ref":"Zig.Nif.Yielding.html","title":"Zig.Nif.Yielding","type":"module"},{"doc":"","ref":"Zig.Nif.Yielding.html#bail/1","title":"Zig.Nif.Yielding.bail/1","type":"function"},{"doc":"","ref":"Zig.Nif.Yielding.html#cancel_join/1","title":"Zig.Nif.Yielding.cancel_join/1","type":"function"},{"doc":"","ref":"Zig.Nif.Yielding.html#frame_cleanup/1","title":"Zig.Nif.Yielding.frame_cleanup/1","type":"function"},{"doc":"","ref":"Zig.Nif.Yielding.html#frame_ptr/1","title":"Zig.Nif.Yielding.frame_ptr/1","type":"function"},{"doc":"","ref":"Zig.Nif.Yielding.html#frame_resources/1","title":"Zig.Nif.Yielding.frame_resources/1","type":"function"},{"doc":"","ref":"Zig.Nif.Yielding.html#frame_type/1","title":"Zig.Nif.Yielding.frame_type/1","type":"function"},{"doc":"","ref":"Zig.Nif.Yielding.html#harness/1","title":"Zig.Nif.Yielding.harness/1","type":"function"},{"doc":"","ref":"Zig.Nif.Yielding.html#harness_fns/1","title":"Zig.Nif.Yielding.harness_fns/1","type":"function"},{"doc":"","ref":"Zig.Nif.Yielding.html#launcher/1","title":"Zig.Nif.Yielding.launcher/1","type":"function"},{"doc":"","ref":"Zig.Nif.Yielding.html#launcher_fns/1","title":"Zig.Nif.Yielding.launcher_fns/1","type":"function"},{"doc":"","ref":"Zig.Nif.Yielding.html#launcher_shim/1","title":"Zig.Nif.Yielding.launcher_shim/1","type":"function"},{"doc":"","ref":"Zig.Nif.Yielding.html#rescheduler/1","title":"Zig.Nif.Yielding.rescheduler/1","type":"function"},{"doc":"the rescheduler fn is a seam between the tail-call reentrancy of the BEAM FFI.","ref":"Zig.Nif.Yielding.html#rescheduler_fn/1","title":"Zig.Nif.Yielding.rescheduler_fn/1","type":"function"},{"doc":"Hooks your zig code into ExUnit, by converting zig tests into ExUnit tests.UsageExampleInside your zig code (dependent.zig):const beam = @import(&quot;beam.zig&quot;); const assert = beam.assert; fn one() i64 { return 1; } test &quot;the one function returns one&quot; { assert(one() == 1); }Inside your elixir code:defmodule MyZigModule do use Zig ~Z&quot;&quot;&quot; const dependent = @import(&quot;dependent.zig&quot;); /// nif: some_nif_fn/1 ... &quot;&quot;&quot; endInside your test module:defmodule MyZigTest do use ExUnit.Case, async: true use Zig.Unit zigtest MyZigModule endScopezigtest will run tests from the following sources:any tests inside of a sigil Z or sigil z constructionany tests inside pub @import zig sources.any tests inside pub usingnamespace zig sources.recursively discovered pub structs.","ref":"Zig.Unit.html","title":"Zig.Unit","type":"module"},{"doc":"loads a module that wraps a zigler NIF, consults the corresponding zig code, generating the corresponding zig tests.Must be called from within a module that has use ExUnit.Case and use Zig","ref":"Zig.Unit.html#zigtest/2","title":"Zig.Unit.zigtest/2","type":"macro"},{"doc":"Fist phase of the Zig compilation process.Looks through the contents of the zig code and creates a map of required source code files. Then outputs a struct containing the relevant information.","ref":"Zig.Assembler.html","title":"Zig.Assembler","type":"module"},{"doc":"","ref":"Zig.Assembler.html#assemble_asset!/2","title":"Zig.Assembler.assemble_asset!/2","type":"function"},{"doc":"assembles zig assets, taking them from their source to and putting them into the target directory.","ref":"Zig.Assembler.html#assemble_assets!/2","title":"Zig.Assembler.assemble_assets!/2","type":"function"},{"doc":"assembles the &quot;core&quot; part of the zig adapter code, these are parts for shimming Zig libraries into the BEAM that are going to be @import'd into the runtime by default.","ref":"Zig.Assembler.html#assemble_kernel!/1","title":"Zig.Assembler.assemble_kernel!/1","type":"function"},{"doc":"","ref":"Zig.Assembler.html#parse_code/2","title":"Zig.Assembler.parse_code/2","type":"function"},{"doc":"","ref":"Zig.Assembler.html#parse_file/2","title":"Zig.Assembler.parse_file/2","type":"function"},{"doc":"","ref":"Zig.Assembler.html#t:file_type/0","title":"Zig.Assembler.file_type/0","type":"type"},{"doc":"","ref":"Zig.Assembler.html#t:t/0","title":"Zig.Assembler.t/0","type":"type"},{"doc":"all code responsible for generating zig code lives in this module.","ref":"Zig.Code.html","title":"Zig.Code","type":"module"},{"doc":"","ref":"Zig.Code.html#adapter/1","title":"Zig.Code.adapter/1","type":"function"},{"doc":"","ref":"Zig.Code.html#aggregate_includes/1","title":"Zig.Code.aggregate_includes/1","type":"function"},{"doc":"","ref":"Zig.Code.html#c_imports/1","title":"Zig.Code.c_imports/1","type":"function"},{"doc":"","ref":"Zig.Code.html#footer/1","title":"Zig.Code.footer/1","type":"function"},{"doc":"","ref":"Zig.Code.html#generate_main/1","title":"Zig.Code.generate_main/1","type":"function"},{"doc":"","ref":"Zig.Code.html#zig_imports/1","title":"Zig.Code.zig_imports/1","type":"function"},{"doc":"contains all parts of the Zig library involved in calling the zig compiler toolchain, especially with regards to the zig command, except for assembling the build.zig file, which is performed by the Zig.Builder module.","ref":"Zig.Command.html","title":"Zig.Command","type":"module"},{"doc":"","ref":"Zig.Command.html#compile/2","title":"Zig.Command.compile/2","type":"function"},{"doc":"","ref":"Zig.Command.html#fetch/1","title":"Zig.Command.fetch/1","type":"function"},{"doc":"","ref":"Zig.Command.html#get_arch/0","title":"Zig.Command.get_arch/0","type":"function"},{"doc":"","ref":"Zig.Command.html#get_os/0","title":"Zig.Command.get_os/0","type":"function"},{"doc":"","ref":"Zig.Command.html#unarchive_zig/1","title":"Zig.Command.unarchive_zig/1","type":"function"},{"doc":"handles instrumenting elixir code with hooks for zig NIFs.","ref":"Zig.Compiler.html","title":"Zig.Compiler","type":"module"},{"doc":"","ref":"Zig.Compiler.html#assembly_dir/2","title":"Zig.Compiler.assembly_dir/2","type":"function"},{"doc":"","ref":"Zig.Compiler.html#function_skeleton/1","title":"Zig.Compiler.function_skeleton/1","type":"function"},{"doc":"","ref":"Zig.Compiler.html#precompile/1","title":"Zig.Compiler.precompile/1","type":"function"},{"doc":"","ref":"Zig.Compiler.html#t:t/0","title":"Zig.Compiler.t/0","type":"type"},{"doc":"handles parsing zig code for the purposes of documentation","ref":"Zig.Doc.Parser.html","title":"Zig.Doc.Parser","type":"module"},{"doc":"","ref":"Zig.Doc.Parser.html#docs_from_dir/2","title":"Zig.Doc.Parser.docs_from_dir/2","type":"function"},{"doc":"Parses the given binary as file_parser.Returns {:ok, [token], rest, context, position, byte_offset} or {:error, reason, rest, context, line, byte_offset} where position describes the location of the file_parser (start position) as {line, column_on_line}.Options:byte_offset - the byte offset for the whole binary, defaults to 0:line - the line and the byte offset into that line, defaults to {1, byte_offset}:context - the initial context value. It will be converted to a map","ref":"Zig.Doc.Parser.html#file_parser/2","title":"Zig.Doc.Parser.file_parser/2","type":"function"},{"doc":"Parses the given binary as head_parser.Returns {:ok, [token], rest, context, position, byte_offset} or {:error, reason, rest, context, line, byte_offset} where position describes the location of the head_parser (start position) as {line, column_on_line}.Options:byte_offset - the byte offset for the whole binary, defaults to 0:line - the line and the byte offset into that line, defaults to {1, byte_offset}:context - the initial context value. It will be converted to a map","ref":"Zig.Doc.Parser.html#head_parser/2","title":"Zig.Doc.Parser.head_parser/2","type":"function"},{"doc":"","ref":"Zig.Doc.Parser.html#moduledocs_from_file/3","title":"Zig.Doc.Parser.moduledocs_from_file/3","type":"function"},{"doc":"decides what features should be documented on zig doc.may be in flux.","ref":"Zig.Doc.Retriever.html","title":"Zig.Doc.Retriever","type":"module"},{"doc":"","ref":"Zig.Doc.Retriever.html#docs_from_dir/2","title":"Zig.Doc.Retriever.docs_from_dir/2","type":"function"},{"doc":"this struct represents all information a zigler module bound to a nif should have. WIP.","ref":"Zig.Module.html","title":"Zig.Module","type":"module"},{"doc":"","ref":"Zig.Module.html#imports/1","title":"Zig.Module.imports/1","type":"function"},{"doc":"","ref":"Zig.Module.html#t:t/0","title":"Zig.Module.t/0","type":"type"},{"doc":"main parsing module: Handles general parsing of nif, resource directives and code elements succeeding those.","ref":"Zig.Parser.html","title":"Zig.Parser","type":"module"},{"doc":"","ref":"Zig.Parser.html#parse/4","title":"Zig.Parser.parse/4","type":"function"},{"doc":"Parses the given binary as parse_zig_block.Returns {:ok, [token], rest, context, position, byte_offset} or {:error, reason, rest, context, line, byte_offset} where position describes the location of the parse_zig_block (start position) as {line, column_on_line}.Options:byte_offset - the byte offset for the whole binary, defaults to 0:line - the line and the byte offset into that line, defaults to {1, byte_offset}:context - the initial context value. It will be converted to a map","ref":"Zig.Parser.html#parse_zig_block/2","title":"Zig.Parser.parse_zig_block/2","type":"function"},{"doc":"","ref":"Zig.Parser.html#t:line_info/0","title":"Zig.Parser.line_info/0","type":"type"},{"doc":"","ref":"Zig.Parser.html#t:parsec_retval/0","title":"Zig.Parser.parsec_retval/0","type":"type"},{"doc":"","ref":"Zig.Parser.html#t:t/0","title":"Zig.Parser.t/0","type":"type"},{"doc":"parses errors emitted by the zig compiler","ref":"Zig.Parser.Error.html","title":"Zig.Parser.Error","type":"module"},{"doc":"given a code file path and a line number, calculates the file and line number of the source document from which it came. Strongly depends on having fencing comments of the form // ref: &lt;file&gt; line: &lt;line&gt; in order to backtrack this information.","ref":"Zig.Parser.Error.html#backreference/2","title":"Zig.Parser.Error.backreference/2","type":"function"},{"doc":"Parses the given binary as check_ref.Returns {:ok, [token], rest, context, position, byte_offset} or {:error, reason, rest, context, line, byte_offset} where position describes the location of the check_ref (start position) as {line, column_on_line}.Options:byte_offset - the byte offset for the whole binary, defaults to 0:line - the line and the byte offset into that line, defaults to {1, byte_offset}:context - the initial context value. It will be converted to a map","ref":"Zig.Parser.Error.html#check_ref/2","title":"Zig.Parser.Error.check_ref/2","type":"function"},{"doc":"given a zig compiler error message, a directory for the code file, and the temporary directory where code assembly is taking place, return an appropriate CompileError struct which can be raised to emit a sensible compiler error message.The temporary directory is stripped (when reasonable) in favor of a &quot;true&quot; source file and any filename substitutions are performed as well.","ref":"Zig.Parser.Error.html#parse/2","title":"Zig.Parser.Error.parse/2","type":"function"},{"doc":"Parses the given binary as parse_error.Returns {:ok, [token], rest, context, position, byte_offset} or {:error, reason, rest, context, line, byte_offset} where position describes the location of the parse_error (start position) as {line, column_on_line}.Options:byte_offset - the byte offset for the whole binary, defaults to 0:line - the line and the byte offset into that line, defaults to {1, byte_offset}:context - the initial context value. It will be converted to a map","ref":"Zig.Parser.Error.html#parse_error/2","title":"Zig.Parser.Error.parse_error/2","type":"function"},{"doc":"For parsing, looking for imports, cimports, and usingnamespace directives.To be completed later","ref":"Zig.Parser.Imports.html","title":"Zig.Parser.Imports","type":"module"},{"doc":"","ref":"Zig.Parser.Imports.html#parse/1","title":"Zig.Parser.Imports.parse/1","type":"function"},{"doc":"Parses the given binary as parse_imports.Returns {:ok, [token], rest, context, position, byte_offset} or {:error, reason, rest, context, line, byte_offset} where position describes the location of the parse_imports (start position) as {line, column_on_line}.Options:byte_offset - the byte offset for the whole binary, defaults to 0:line - the line and the byte offset into that line, defaults to {1, byte_offset}:context - the initial context value. It will be converted to a map","ref":"Zig.Parser.Imports.html#parse_imports/2","title":"Zig.Parser.Imports.parse_imports/2","type":"function"},{"doc":"","ref":"Zig.Parser.Imports.html#t:line_info/0","title":"Zig.Parser.Imports.line_info/0","type":"type"},{"doc":"","ref":"Zig.Parser.Imports.html#t:parsec_retval/0","title":"Zig.Parser.Imports.parsec_retval/0","type":"type"},{"doc":"","ref":"Zig.Parser.Imports.html#t:t/0","title":"Zig.Parser.Imports.t/0","type":"type"},{"doc":"This datastructure represents structured information about a single nif inside of a Zig.sigil_Z/2 block. This is used to generate the exported_nifs variable which is an array of ErlNifFunc structs. The following keys are implemented:name: (atom/0) the function name to be bound into the modulearity: (arity/0) the arity of the erlang function (the zig function may have a different arity).doc: (iodata/0) zig docstrings which should be turned into elixir docsargs: (String.t/0) a list of zig types which are the arguments for the functionretval: (String.t/0) the type of the return valueopts: (t:keyword) list of nif options. These options are currently supported:concurrency: &lt;model&gt; picks a long-running nif concurrency model. The following concurrency models are supported::threaded -- if the nif should run in a separate OS thread.:yielding -- if the nif should use zig's yield keyword to yield to the BEAM scheduler.:dirty_cpu -- if the nif should run in a dirty cpu scheduler.:dirty_io -- if the nif should run in a dirty io scheduler.","ref":"Zig.Parser.Nif.html","title":"Zig.Parser.Nif","type":"module"},{"doc":"","ref":"Zig.Parser.Nif.html#register_function_header/2","title":"Zig.Parser.Nif.register_function_header/2","type":"function"},{"doc":"","ref":"Zig.Parser.Nif.html#validate_args/3","title":"Zig.Parser.Nif.validate_args/3","type":"function"},{"doc":"","ref":"Zig.Parser.Nif.html#validate_arity/3","title":"Zig.Parser.Nif.validate_arity/3","type":"function"},{"doc":"","ref":"Zig.Parser.Nif.html#validate_retval/3","title":"Zig.Parser.Nif.validate_retval/3","type":"function"},{"doc":"","ref":"Zig.Parser.Nif.html#t:concurrency/0","title":"Zig.Parser.Nif.concurrency/0","type":"type"},{"doc":"","ref":"Zig.Parser.Nif.html#t:option/0","title":"Zig.Parser.Nif.option/0","type":"type"},{"doc":"","ref":"Zig.Parser.Nif.html#t:t/0","title":"Zig.Parser.Nif.t/0","type":"type"},{"doc":"resource parser object","ref":"Zig.Parser.Resource.html","title":"Zig.Parser.Resource","type":"module"},{"doc":"","ref":"Zig.Parser.Resource.html#register_resource_definition/1","title":"Zig.Parser.Resource.register_resource_definition/1","type":"function"},{"doc":"","ref":"Zig.Parser.Resource.html#t:t/0","title":"Zig.Parser.Resource.t/0","type":"type"},{"doc":"resource cleanup struct","ref":"Zig.Parser.ResourceCleanup.html","title":"Zig.Parser.ResourceCleanup","type":"module"},{"doc":"","ref":"Zig.Parser.ResourceCleanup.html#register_function_header/2","title":"Zig.Parser.ResourceCleanup.register_function_header/2","type":"function"},{"doc":"","ref":"Zig.Parser.ResourceCleanup.html#validate_args/3","title":"Zig.Parser.ResourceCleanup.validate_args/3","type":"function"},{"doc":"","ref":"Zig.Parser.ResourceCleanup.html#validate_arity/3","title":"Zig.Parser.ResourceCleanup.validate_arity/3","type":"function"},{"doc":"","ref":"Zig.Parser.ResourceCleanup.html#validate_retval/3","title":"Zig.Parser.ResourceCleanup.validate_retval/3","type":"function"},{"doc":"","ref":"Zig.Parser.ResourceCleanup.html#t:t/0","title":"Zig.Parser.ResourceCleanup.t/0","type":"type"},{"doc":"parses zig code and converts test blocks to test functions","ref":"Zig.Parser.Unit.html","title":"Zig.Parser.Unit","type":"module"},{"doc":"","ref":"Zig.Parser.Unit.html#parse/2","title":"Zig.Parser.Unit.parse/2","type":"function"},{"doc":"Parses the given binary as unit_parser.Returns {:ok, [token], rest, context, position, byte_offset} or {:error, reason, rest, context, line, byte_offset} where position describes the location of the unit_parser (start position) as {line, column_on_line}.Options:byte_offset - the byte offset for the whole binary, defaults to 0:line - the line and the byte offset into that line, defaults to {1, byte_offset}:context - the initial context value. It will be converted to a map","ref":"Zig.Parser.Unit.html#unit_parser/2","title":"Zig.Parser.Unit.unit_parser/2","type":"function"},{"doc":"occasionally zig and/or the llvm bits that get pulled down have errors. This module actively patches code when this situation occurs.","ref":"Zig.Patches.html","title":"Zig.Patches","type":"module"},{"doc":"","ref":"Zig.Patches.html#patch_needed?/2","title":"Zig.Patches.patch_needed?/2","type":"function"},{"doc":"","ref":"Zig.Patches.html#sync/1","title":"Zig.Patches.sync/1","type":"function"},{"doc":"handles creating typespecs for zig Nif functions","ref":"Zig.Typespec.html","title":"Zig.Typespec","type":"module"},{"doc":"general-purpose BEAM mutex wrapped into the zig mutex interface","ref":"beam_mutex.html","title":"beam_mutex","type":"module"},{"doc":"initializes the mutex. Note this is failable.","ref":"beam_mutex.html#init/1","title":"beam_mutex.init/1","type":"function"},{"doc":"Try to acquire the mutex without blocking. Returns null if the mutex is unavailable. Otherwise returns Held. Call release on Held.","ref":"beam_mutex.html#tryAcquire/1","title":"beam_mutex.tryAcquire/1","type":"function"},{"doc":"Acquire the mutex. Will deadlock if the mutex is already held by the calling thread.","ref":"beam_mutex.html#acquire/1","title":"beam_mutex.acquire/1","type":"function"},{"doc":"Provides erlang's NIF convenience functions to your Zig code.This struct derives from zig/beam/erl_nif.zig. Usually brought in as const e = @import(&quot;erl_nif.zig&quot;).c;, leading to the aliased shortcut e.for example, to call the enif_alloc function, you would use the following code:const e = @import(&quot;erl_nif.zig&quot;).c; pub fn give_me_ten_bytes() ?*u8 { return e.enif_alloc(10); }refer to the erlang documentation for available functions","ref":"erl_nif.html","title":"erl_nif","type":"module"},{"doc":"This struct contains adapters designed to facilitate interfacing the BEAM's c-style helpers for NIFs with a more idiomatic Zig-style of programming, for example, the use of slices instead of null-terminated arrays as strings.This struct derives from zig/beam/beam.zig, and you may import it into your module's zig code by calling:const beam = @import(&quot;beam.zig&quot;)This is done automatically for you inside your ~Z forms, so do NOT use this import statement with inline Zig.FeaturesThe BEAM AllocatorWraps e.enif_alloc and e.enif_free functions into a compliant Zig allocator struct. You should thus be able to supply Zig standard library functions which require an allocator a struct that is compliant with its requirements.This is, in particular, useful for slice generation.Example (slice generation)beam = @import(&quot;beam.zig&quot;); fn make_a_slice_of_floats() ![]f32 { return beam.allocator.alloc(f32, 100); }Beacuse Zig features composable allocators, you can very easily implement custom allocators on top of the existing BEAM allocator.GettersErlang's NIF interface provides a comprehensive set of methods to retrieve data out of BEAM terms. However, this set of methods presents an error handling scheme that is designed for C and inconsistent with the idiomatic scheme used for Zig best practices.A series of get functions is provided, implementing these methods in accordance to best practices. These include get/3, which is the generic method for getting scalar values, get_X, which are typed methods for retrieving scalar values, and get_slice_of/3, which is the generic method for retrieving a Zig slice from a BEAM list.Naturally, for all of these functions, you will have to provide the BEAM environment value.Examplesconst beam = @import(&quot;beam.zig&quot;); fn double_value(env: beam.env, value: beam.term) !f64 { return (try beam.get_f64(env, value)) * 2; } fn sum_float_list(env: beam.env, list: beam.term) !f64 { zig_list: []f64 = try beam.get_slice_of(f64, env, list); defer beam.allocator.free(zig_list); // don&#39;t forget to clean up! result: f64 = 0; for (list) |item| { result += item; } return result; }MakersA series of &quot;make&quot; functions is provided which allow for easy export of Zig values back to the BEAM. Typically, these functions are used in the automatic type marshalling performed by Zigler, however, you may want to be able to use them yourself to assemble BEAM datatypes not directly supported by Zig. For example, a custom tuple value.Exampleconst beam = @import(&quot;beam.zig&quot;); const ok_slice=&quot;ok&quot;[0..]; fn to_ok_tuple(env: beam.env, value: i64) !beam.term { var tuple_slice: []term = try beam.allocator.alloc(beam.term, 2); defer beam.allocator.free(tuple_slice); tuple_slice[0] = beam.make_atom(env, ok_slice); tuple_slice[1] = beam.make_i64(env, value); return beam.make_tuple(env, tuple_slice); }","ref":"beam.html","title":"beam","type":"module"},{"doc":"A helper for marshalling values from the BEAM runtime into Zig. Use this function if you need support for Zig generics.Used internally to typcheck values coming into Zig slice.supported types:c_intc_longisizeusizeu8i32i64f16f32f64","ref":"beam.html#get/3","title":"beam.get/3","type":"function"},{"doc":"Takes a BEAM int term and returns a c_int value. Should only be used for C interop with Zig functions.Raises beam.Error.FunctionClauseError if the term is not integer/0","ref":"beam.html#get_c_int/2","title":"beam.get_c_int/2","type":"function"},{"doc":"Takes a BEAM int term and returns a c_uint value. Should only be used for C interop with Zig functions.Raises beam.Error.FunctionClauseError if the term is not integer/0","ref":"beam.html#get_c_uint/2","title":"beam.get_c_uint/2","type":"function"},{"doc":"Takes a BEAM int term and returns a c_long value. Should only be used for C interop with Zig functions.Raises beam.Error.FunctionClauseError if the term is not integer/0","ref":"beam.html#get_c_long/2","title":"beam.get_c_long/2","type":"function"},{"doc":"Takes a BEAM int term and returns a c_ulong value. Should only be used for C interop with Zig functions.Raises beam.Error.FunctionClauseError if the term is not integer/0","ref":"beam.html#get_c_ulong/2","title":"beam.get_c_ulong/2","type":"function"},{"doc":"Takes a BEAM int term and returns a isize value. Should only be used for C interop.Raises beam.Error.FunctionClauseError if the term is not integer/0","ref":"beam.html#get_isize/2","title":"beam.get_isize/2","type":"function"},{"doc":"Takes a BEAM int term and returns a usize value. Zig idiomatically uses usize for its size values, so typically you should be using this function.Raises beam.Error.FunctionClauseError if the term is not integer/0","ref":"beam.html#get_usize/2","title":"beam.get_usize/2","type":"function"},{"doc":"Takes a BEAM int term and returns a u8 value.Note that this conversion function checks to make sure it's in range (0..255).Raises beam.Error.FunctionClauseError if the term is not integer/0","ref":"beam.html#get_u8/2","title":"beam.get_u8/2","type":"function"},{"doc":"Takes a BEAM int term and returns a u16 value.Note that this conversion function checks to make sure it's in range (0..65535).Raises beam.Error.FunctionClauseError if the term is not integer/0","ref":"beam.html#get_u16/2","title":"beam.get_u16/2","type":"function"},{"doc":"Takes a BEAM int term and returns a u32 value.Raises beam.Error.FunctionClauseError if the term is not integer/0","ref":"beam.html#get_u32/2","title":"beam.get_u32/2","type":"function"},{"doc":"Takes a BEAM int term and returns a u64 value.Raises beam.Error.FunctionClauseError if the term is not integer/0","ref":"beam.html#get_u64/2","title":"beam.get_u64/2","type":"function"},{"doc":"Takes a BEAM int term and returns an i32 value.Note that this conversion function does not currently do range checking.Raises beam.Error.FunctionClauseError if the term is not integer/0","ref":"beam.html#get_i32/2","title":"beam.get_i32/2","type":"function"},{"doc":"Takes a BEAM int term and returns an i64 value.Note that this conversion function does not currently do range checking.Raises beam.Error.FunctionClauseError if the term is not integer/0","ref":"beam.html#get_i64/2","title":"beam.get_i64/2","type":"function"},{"doc":"Takes a BEAM float term and returns an f16 value.Note that this conversion function does not currently do range checking.Raises beam.Error.FunctionClauseError if the term is not float/0","ref":"beam.html#get_f16/2","title":"beam.get_f16/2","type":"function"},{"doc":"Takes a BEAM float term and returns an f32 value.Note that this conversion function does not currently do range checking.Raises beam.Error.FunctionClauseError if the term is not float/0","ref":"beam.html#get_f32/2","title":"beam.get_f32/2","type":"function"},{"doc":"Takes a BEAM float term and returns an f64 value.Raises beam.Error.FunctionClauseError if the term is not float/0","ref":"beam.html#get_f64/2","title":"beam.get_f64/2","type":"function"},{"doc":"Takes a BEAM atom term and retrieves it as a slice []u8 value. it's the caller's responsibility to make sure that the value is freed.Uses the standard beam.allocator allocator. If you require a custom allocator, use get_atom_slice_alloc/3Raises beam.Error.FunctionClauseError if the term is not atom/0","ref":"beam.html#get_atom_slice/2","title":"beam.get_atom_slice/2","type":"function"},{"doc":"Takes a BEAM atom term and retrieves it as a slice []u8 value, with any allocator.Raises beam.Error.FunctionClauseError if the term is not atom/0","ref":"beam.html#get_atom_slice_alloc/3","title":"beam.get_atom_slice_alloc/3","type":"function"},{"doc":"Takes an BEAM binary/0 term and retrieves a pointer to the binary data as a Zig c-string ([*c]u8). No memory is allocated for this operation.Should only be used for c interop functions.Note: this function could have unexpected results if your BEAM binary contains any zero byte values. Always use get_char_slice/2 when C-interop is not necessary.Raises beam.Error.FunctionClauseError if the term is not binary/0","ref":"beam.html#get_c_string/2","title":"beam.get_c_string/2","type":"function"},{"doc":"Takes an BEAM binary/0 term and retrieves it as a Zig character slice ([]u8) No memory is allocated for this operation.Raises beam.Error.FunctionClauseError if the term is not binary/0","ref":"beam.html#get_char_slice/2","title":"beam.get_char_slice/2","type":"function"},{"doc":"Takes an BEAM binary/0 term and returns the corresponding binary struct.Raises beam.Error.FunctionClauseError if the term is not binary/0","ref":"beam.html#get_binary/2","title":"beam.get_binary/2","type":"function"},{"doc":"Takes an BEAM pid/0 term and returns the corresponding pid struct.Note that this is a fairly opaque struct and you're on your own as to what you can do with this (for now), except as a argument for the e.enif_send function.Raises beam.Error.FunctionClauseError if the term is not pid/0","ref":"beam.html#get_pid/2","title":"beam.get_pid/2","type":"function"},{"doc":"shortcut for e.enif_self, marshalling into zig error style.returns the pid value if it's env is a process-bound environment, otherwise returns beam.Error.FunctionClauseError.","ref":"beam.html#self/1","title":"beam.self/1","type":"function"},{"doc":"shortcut for e.enif_selfreturns true if the send is successful, false otherwise.NOTE this function assumes a valid BEAM environment. If you have spawned an OS thread without a BEAM environment, you must use send_advanced/4","ref":"beam.html#send/3","title":"beam.send/3","type":"function"},{"doc":"shortcut for e.enif_selfreturns true if the send is successful, false otherwise.if you are sending from a thread that does not have a BEAM environment, you should put null in both environment variables.","ref":"beam.html#send_advanced/4","title":"beam.send_advanced/4","type":"function"},{"doc":"Takes an Beam tuple/0 term and returns it as a slice of term structs. Does not allocate memory for this operation.Raises beam.Error.FunctionClauseError if the term is not tuple/0","ref":"beam.html#get_tuple/2","title":"beam.get_tuple/2","type":"function"},{"doc":"Takes a BEAM list/0 term and returns its length.Raises beam.Error.FunctionClauseError if the term is not list/0","ref":"beam.html#get_list_length/2","title":"beam.get_list_length/2","type":"function"},{"doc":"Iterates over a BEAM list/0.In this function, the list value will be modified to the tl of the BEAM list, and the return value will be the BEAM term.Raises beam.Error.FunctionClauseError if the term is not list/0","ref":"beam.html#get_head_and_iter/2","title":"beam.get_head_and_iter/2","type":"function"},{"doc":"A generic function which lets you convert a BEAM list/0 of homogeous type into a Zig slice.The resulting slice will be allocated using the beam allocator, with ownership passed to the caller. If you need to use a different allocator, use get_slice_of_alloc/4Raises beam.Error.FunctionClauseError if the term is not list/0. Also raises beam.Error.FunctionClauseError if any of the terms is incompatible with the internal typesupported internal types:c_intc_longisizeusizeu8i32i64f16f32f64","ref":"beam.html#get_slice_of/3","title":"beam.get_slice_of/3","type":"function"},{"doc":"Converts an BEAM list/0 of homogenous type into a Zig slice, but using any allocator you wish.ownership is passed to the caller.Raises beam.Error.FunctionClauseError if the term is not list/0. Also raises beam.Error.FunctionClauseError if any of the terms is incompatible with the internal type.supported internal types:c_intc_longisizeusizeu8i32i64f16f32f64","ref":"beam.html#get_slice_of_alloc/4","title":"beam.get_slice_of_alloc/4","type":"function"},{"doc":"Converts an BEAM boolean/0 into a Zig bool.Raises beam.Error.FunctionClauseError if the term is not boolean/0. May potentially raise an out of memory error, as it must make an allocation to perform its conversion.","ref":"beam.html#get_bool/2","title":"beam.get_bool/2","type":"function"},{"doc":"A helper for marshalling values from Zig back into the runtime. Use this function if you need support for Zig generics.supported types:c_intc_longisizeusizeu8i32i64f16f32f64","ref":"beam.html#make/3","title":"beam.make/3","type":"function"},{"doc":"converts a char (u8) value into a BEAM integer/0.","ref":"beam.html#make_u8/2","title":"beam.make_u8/2","type":"function"},{"doc":"converts a unsigned (u16) value into a BEAM integer/0.","ref":"beam.html#make_u16/2","title":"beam.make_u16/2","type":"function"},{"doc":"converts a unsigned (u32) value into a BEAM integer/0.","ref":"beam.html#make_u32/2","title":"beam.make_u32/2","type":"function"},{"doc":"converts a unsigned (u64) value into a BEAM integer/0.","ref":"beam.html#make_u64/2","title":"beam.make_u64/2","type":"function"},{"doc":"converts a c_int value into a BEAM integer/0.","ref":"beam.html#make_c_int/2","title":"beam.make_c_int/2","type":"function"},{"doc":"converts a c_uint value into a BEAM integer/0.","ref":"beam.html#make_c_uint/2","title":"beam.make_c_uint/2","type":"function"},{"doc":"converts a c_long value into a BEAM integer/0.","ref":"beam.html#make_c_long/2","title":"beam.make_c_long/2","type":"function"},{"doc":"converts a c_ulong value into a BEAM integer/0.","ref":"beam.html#make_c_ulong/2","title":"beam.make_c_ulong/2","type":"function"},{"doc":"converts an isize value into a BEAM integer/0.","ref":"beam.html#make_isize/2","title":"beam.make_isize/2","type":"function"},{"doc":"converts a usize value into a BEAM integer/0.","ref":"beam.html#make_usize/2","title":"beam.make_usize/2","type":"function"},{"doc":"converts an i32 value into a BEAM integer/0.","ref":"beam.html#make_i32/2","title":"beam.make_i32/2","type":"function"},{"doc":"converts an i64 value into a BEAM integer/0.","ref":"beam.html#make_i64/2","title":"beam.make_i64/2","type":"function"},{"doc":"converts an f16 value into a BEAM float/0.","ref":"beam.html#make_f16/2","title":"beam.make_f16/2","type":"function"},{"doc":"converts an f32 value into a BEAM float/0.","ref":"beam.html#make_f32/2","title":"beam.make_f32/2","type":"function"},{"doc":"converts an f64 value into a BEAM float/0.","ref":"beam.html#make_f64/2","title":"beam.make_f64/2","type":"function"},{"doc":"converts a Zig char slice ([]u8) into a BEAM atom/0.","ref":"beam.html#make_atom/2","title":"beam.make_atom/2","type":"function"},{"doc":"converts a Zig char slice ([]u8) into a BEAM binary/0.no memory allocation inside of Zig is performed and the BEAM environment is responsible for the resulting binary. You are responsible for managing the allocation of the slice.","ref":"beam.html#make_slice/2","title":"beam.make_slice/2","type":"function"},{"doc":"converts an c string ([*c]u8) into a BEAM binary/0. Mostly used for c interop.no memory allocation inside of Zig is performed and the BEAM environment is responsible for the resulting binary. You are responsible for managing the allocation of the slice.","ref":"beam.html#make_c_string/2","title":"beam.make_c_string/2","type":"function"},{"doc":"converts a slice of terms into a BEAM tuple/0.","ref":"beam.html#make_tuple/2","title":"beam.make_tuple/2","type":"function"},{"doc":"converts a slice of terms into a BEAM list/0.","ref":"beam.html#make_term_list/2","title":"beam.make_term_list/2","type":"function"},{"doc":"converts a Zig char slice ([]u8) into a BEAM charlist/0.","ref":"beam.html#make_charlist/2","title":"beam.make_charlist/2","type":"function"},{"doc":"converts a c string ([*c]u8) into a BEAM charlist/0.","ref":"beam.html#make_cstring_charlist/2","title":"beam.make_cstring_charlist/2","type":"function"},{"doc":"A helper to make BEAM lists out of slices of term. Use this function if you need a generic listbuilding function.uses the BEAM allocator internally. If you would like to use a custom allocator, (for example an arena allocator, if you have very long lists), use make_list_alloc/4supported internal types:c_intc_longisizeusizeu8i32i64f16f32f64","ref":"beam.html#make_list/3","title":"beam.make_list/3","type":"function"},{"doc":"A helper to make a BEAM t:Kernel.list out of terms, with any allocator. Use this function if you need a generic listbuilding function.supported internal types:c_intc_longisizeusizeu8i32i64f16f32f64","ref":"beam.html#make_list_alloc/4","title":"beam.make_list_alloc/4","type":"function"},{"doc":"converts a c_int slice ([]c_int) into a BEAM list of integer/0.","ref":"beam.html#make_c_int_list/2","title":"beam.make_c_int_list/2","type":"function"},{"doc":"converts a c_long slice ([]c_long) into a BEAM list of integer/0.","ref":"beam.html#make_c_long_list/2","title":"beam.make_c_long_list/2","type":"function"},{"doc":"converts an i32 slice ([]i32) into a BEAM list of integer/0.","ref":"beam.html#make_i32_list/2","title":"beam.make_i32_list/2","type":"function"},{"doc":"converts an i64 slice ([]i64) into a BEAM list of integer/0.","ref":"beam.html#make_i64_list/2","title":"beam.make_i64_list/2","type":"function"},{"doc":"converts an f16 slice ([]f16) into a BEAM list of float/0.","ref":"beam.html#make_f16_list/2","title":"beam.make_f16_list/2","type":"function"},{"doc":"converts an f32 slice ([]f32) into a BEAM list of float/0.","ref":"beam.html#make_f32_list/2","title":"beam.make_f32_list/2","type":"function"},{"doc":"converts an f64 slice ([]f64) into a BEAM list of float/0.","ref":"beam.html#make_f64_list/2","title":"beam.make_f64_list/2","type":"function"},{"doc":"converts a bool value into a boolean/0 value.","ref":"beam.html#make_bool/2","title":"beam.make_bool/2","type":"function"},{"doc":"creates a beam nil value.","ref":"beam.html#make_nil/1","title":"beam.make_nil/1","type":"function"},{"doc":"creates a beam ok value.","ref":"beam.html#make_ok/1","title":"beam.make_ok/1","type":"function"},{"doc":"creates a beam error value.","ref":"beam.html#make_error/1","title":"beam.make_error/1","type":"function"},{"doc":"A helper to make {:ok, term} terms from arbitrarily-typed values.supported types:c_intc_longisizeusizeu8i32i64f16f32f64Use make_ok_term/2 to make ok tuples from generic terms. Use make_ok_atom/2 to make ok tuples with atom terms from slices.","ref":"beam.html#make_ok_tuple/3","title":"beam.make_ok_tuple/3","type":"function"},{"doc":"A helper to make {:ok, binary} terms from slices","ref":"beam.html#make_ok_binary/2","title":"beam.make_ok_binary/2","type":"function"},{"doc":"A helper to make {:ok, atom} terms from slices","ref":"beam.html#make_ok_atom/2","title":"beam.make_ok_atom/2","type":"function"},{"doc":"A helper to make {:ok, term} terms in general","ref":"beam.html#make_ok_term/2","title":"beam.make_ok_term/2","type":"function"},{"doc":"A helper to make {:error, term} terms from arbitrarily-typed values.supported types:c_intc_longisizeusizeu8i32i64f16f32f64Use make_error_term/2 to make error tuples from generic terms. Use make_error_atom/2 to make atom errors from slices.","ref":"beam.html#make_error_tuple/3","title":"beam.make_error_tuple/3","type":"function"},{"doc":"A helper to make {:error, atom} terms from slices","ref":"beam.html#make_error_atom/2","title":"beam.make_error_atom/2","type":"function"},{"doc":"A helper to make {:error, binary} terms from slices","ref":"beam.html#make_error_binary/2","title":"beam.make_error_binary/2","type":"function"},{"doc":"A helper to make {:error, term} terms in general","ref":"beam.html#make_error_term/2","title":"beam.make_error_term/2","type":"function"},{"doc":"Encapsulates e.enif_make_ref and allows it to return a FunctionClauseError.","ref":"beam.html#make_ref/1","title":"beam.make_ref/1","type":"function"},{"doc":"this function is going to be dropped inside the suspend statement.","ref":"beam.html#yield/0","title":"beam.yield/0","type":"function"},{"doc":"This function is used to communicate :enomem back to the BEAM as an exception.The BEAM is potentially OOM-safe, and Zig lets you leverage that. OOM errors from beam.allocator can be converted to a generic erlang term that represents an exception. Returning this from your NIF results in a BEAM throw event.","ref":"beam.html#raise_enomem/1","title":"beam.raise_enomem/1","type":"function"},{"doc":"This function is used to communicate :function_clause back to the BEAM as an exception.By default Zigler will do argument input checking on value ingress from the dynamic BEAM runtime to the static Zig runtime. You can also use this function to communicate a similar error by returning the resulting term from your NIF.","ref":"beam.html#raise_function_clause_error/1","title":"beam.raise_function_clause_error/1","type":"function"},{"doc":"This function is used to communicate :resource_error back to the BEAM as an exception.","ref":"beam.html#raise_resource_error/1","title":"beam.raise_resource_error/1","type":"function"},{"doc":"This function is used to communicate :assertion_error back to the BEAM as an exception.Used when running Zigtests, when trapping beam.AssertionError.AssertionError.","ref":"beam.html#raise_assertion_error/1","title":"beam.raise_assertion_error/1","type":"function"},{"doc":"A function used to return assertion errors to a zigtest.Zig's std.assert() will panic the Zig runtime and therefore the entire BEAM VM, making it incompatible with Elixir's Unit tests. As the VM is required for certain functionality (like e.enif_alloc), a BEAM-compatible assert is necessary.When building zigtests, assert(...) calls get lexically converted to try beam.assert(...) calls.","ref":"beam.html#assert/3","title":"beam.assert/3","type":"function"},{"doc":"Translates to ExUnit.AssertionError. Mostly used in Zig unit tests.All test clauses in the directories of your Zig-enabled modules are converted to Zig functions with the inferred type !void. The beam.assert/1 function can throw this error as its error type.Zigler converts assert statements in test blocks to try beam.assert(...);","ref":"beam.AssertionError.AssertionError.html","title":"beam.AssertionError.AssertionError","type":"exception"},{"doc":"Occurs when there's a problem launching a threaded nif.","ref":"beam.ThreadError.LaunchError.html","title":"beam.ThreadError.LaunchError","type":"exception"},{"doc":"ZiglerLibrary test status:","ref":"readme.html","title":"Zigler","type":"extras"},{"doc":"Zigler is available in Hex, and the package can be installed by adding zigler to your list of dependencies in mix.exs:def deps do [ {:zigler, &quot;~&gt; 0.7.2&quot;, runtime: false} ] end","ref":"readme.html#installation","title":"Zigler - Installation","type":"extras"},{"doc":"Docs can be found at https://hexdocs.pm/zigler.","ref":"readme.html#documentation","title":"Zigler - Documentation","type":"extras"},{"doc":"LinuxFreeBSD (tested, but not subjected to CI)MacOS (I believe it works but is still offically untested)Nerves cross-compilation is supported out of the box.","ref":"readme.html#currently-supported-platforms","title":"Zigler - Currently supported platforms","type":"extras"},{"doc":"Wouldn't it be nice if you could make NIFs as easily as you can use the asm keyword in C?This is now possible, using the magic of Zig.defmodule ExampleZig do use Zig ~Z&quot;&quot;&quot; /// nif: example_fun/2 fn example_fun(value1: f64, value2: f64) bool { return value1 &gt; value2; } &quot;&quot;&quot; end test &quot;example nifs&quot; do assert ExampleZig.example_fun(0.8, -0.8) refute ExampleZig.example_fun(0.1, 0.4) endZigler will do automatic type marshalling between Elixir code and Zig code. It will also convert trickier types into types you care about, for example:defmodule ZigCollections do use Zig ~Z&quot;&quot;&quot; /// nif: string_count/1 fn string_count(string: []u8) i64 { return @intCast(i64, string.len); } /// nif: list_sum/1 fn list_sum(array: []f64) f64 { var sum: f64 = 0.0; for(array) | item | { sum += item; } return sum; } &quot;&quot;&quot; end test &quot;type marshalling&quot; do assert 9 == ZigCollections.string_count(&quot;hello zig&quot;) assert 6.0 == ZigCollections.list_sum([1.0, 2.0, 3.0]) endMemory allocation with zigler is easy! A standard BEAM allocator is provided for you, so any zig code you import will play nice with the BEAM.defmodule Allocations do use Zig ~Z&quot;&quot;&quot; /// nif: double_atom/1 fn double_atom(env: beam.env, string: []u8) beam.term { var double_string = beam.allocator.alloc(u8, string.len * 2) catch { return beam.raise_enomem(env); }; defer beam.allocator.free(double_string); for (string) | char, i | { double_string[i] = char; double_string[i + string.len] = char; } return beam.make_atom(env, double_string); } &quot;&quot;&quot; end test &quot;allocations&quot; do assert :foofoo == Allocations.double_atom(&quot;foo&quot;) endIt is a goal for Zigler to make using it to bind C libraries easier than using C to bind C libraries. Here is an example:defmodule BlasDynamic do use Zig, libs: [&quot;/usr/lib/x86_64-linux-gnu/blas/libblas.so&quot;], include: [&quot;/usr/include/x86_64-linux-gnu&quot;], link_libc: true ~Z&quot;&quot;&quot; const blas = @cImport({ @cInclude(&quot;cblas.h&quot;); }); /// nif: blas_axpy/3 fn blas_axpy(env: beam.env, a: f64, x: []f64, y: []f64) beam.term { if (x.len != y.len) { return beam.raise_function_clause_error(env); } blas.cblas_daxpy(@intCast(c_int, x.len), a, x.ptr, 1, y.ptr, 1); return beam.make_f64_list(env, y) catch { return beam.raise_function_clause_error(env); }; } &quot;&quot;&quot; end test &quot;we can use a blas shared library&quot; do # returns aX+Y assert [11.0, 18.0] == BlasDynamic.blas_axpy(3.0, [2.0, 4.0], [5.0, 6.0]) endDocumentationYou can document nif functions, local functions, zig structs, variables, and types. If you document a nif function, it will be a part of the module documentation, and accessible using the iex h method, etc.Example:defmodule Documentation do use Zig ~Z&quot;&quot;&quot; /// a zero-arity function which returns 47. /// nif: zero_arity/0 fn zero_arity() i64 { return 47; } &quot;&quot;&quot; end","ref":"readme.html#zig-nifs-made-easy","title":"Zigler - Zig Nifs made easy","type":"extras"},{"doc":"Make being a good citizen of the BEAM easy.Use magic, but sparingly, only to prevent errors.Let the user see behind the curtain.Let the user opt out of magic.Magic shouldn't get in the way.","ref":"readme.html#zigler-principles","title":"Zigler - Zigler Principles","type":"extras"},{"doc":"Using NifsNifs are the entrypoint between your BEAM code and your C ABI code. Zigler provides semantics which are designed to make it easy to write safe NIF code with the safety, memory guarantees that Zig provides.","ref":"nifs.html","title":"Using Nifs","type":"extras"},{"doc":"~Z&quot;&quot;&quot; /// nif: add_one/1 fn add_one(input: i32) i32 { return input + 1; } &quot;&quot;&quot; test &quot;add one&quot; do assert 48 == add_one(47) endsimply define a zig fn with desired input/output parameters. Zigler will generate a corresponding function in the surrounding Elixir module, and a mismatched value will raise a FunctionClauseError. The following types are accepted as and outputs:boolu8u16i32u32i64u64c_int *c_uint *c_long *c_ulong *isize *usizebeam.termbeam.pid*only use these types to interface with C code and libraries.Slices of all of the above, except u8, are also accepted (see below).Example: Slices~Z&quot;&quot;&quot; /// nif: sum/1 fn sum(input: []f32) f32 { var total: f32 = 0.0; for (input) | value | { total += value; } return total; } &quot;&quot;&quot; test &quot;sum&quot; do assert 6.0 == sum([1.0, 2.0, 3.0]) endExample: BinariesBinaries are marshalled into u8 slices.~Z&quot;&quot;&quot; /// nif: capitalize/1 fn capitalize(input: []u8) []u8 { // note this is poorly bounds tested. input[0] = input[0] - 32; return input; } &quot;&quot;&quot; test &quot;capitalize&quot; do assert &quot;Foo&quot; == capitalize(&quot;foo&quot;) endExample: FunctionClauseErrortest &quot;functionclauseerror&quot; do assert_raise FunctionClauseError, fn -&gt; capitalize(:atom) end end","ref":"nifs.html#basic-operation","title":"Using Nifs - Basic operation","type":"extras"},{"doc":"In order to build complex terms to be consumed by the rest of the BEAM, you need to have access to the &quot;environment&quot; of the calling process. In order to be provided this, you may request it as the first parameter to your zig function. Play close attention to the arity of the nif declaration.~Z&quot;&quot;&quot; /// nif: add_three/1 fn add_three(env: beam.env, number: i32) beam.term { return beam.make_ok_term(env, beam.make_i32(env, number + 3)); } &quot;&quot;&quot; test &quot;add_three&quot; do assert {:ok, 6} = add_three(3) end","ref":"nifs.html#optional-beam-env-term","title":"Using Nifs - Optional beam.env term","type":"extras"},{"doc":"Use the beam.allocator to allocate memory when you need it. The allocator conforms to the Zig allocator interface and importantly communicates allocations back to the BEAM for tracking.~Z&quot;&quot;&quot; // note: this function deliberately leaks memory!! /// nif: allocate_leak/0 fn allocate_leak(env: beam.env) beam.term { _ = beam.allocator.alloc(u8, 10_000_000) catch return beam.raise_enomem(env); return beam.make_ok(env); } &quot;&quot;&quot; test &quot;allocate_leak&quot; do start_memory = :erlang.memory[:total] assert :ok == allocate_leak() after_memory = :erlang.memory[:total] # evidence that we have leaked memory! assert 8_000_000 &lt; (after_memory - start_memory) end","ref":"nifs.html#allocation","title":"Using Nifs - Allocation","type":"extras"},{"doc":"If you want to launch your nif as a yielding, use the yielding attribute. This will allow the BEAM to schedule around the yield points. This is the most desirable way of creating a long-running NIF (when it gets repaired).it is safe to use the beam.yield() function in non-yielding functions. you can thus write a yielding NIF function and then test the performance delta with, for example dirty or threaded NIFs and decide if behaving correctly is worth the performance cost.if the running process is stopped by the VM, the nif will be caught by the cleanup method and will return beam.YieldError. You must catch this error and return to allow the nif to be awaited.future editions of this feature will let you forward the yield-error to the cleanup method so you can use try instead of catch.~Z&quot;&quot;&quot; /// nif: yielding/0 yielding fn yielding(env: beam.env) beam.term { var count: i32 = 0; while (count &lt; 10_000) { // // do some work here // // yield point beam.yield() catch return beam.make_error(env); count += 1; } return beam.make_ok(env); } &quot;&quot;&quot; test &quot;yielding&quot; do assert :ok = yielding() end","ref":"nifs.html#yielding-nifs","title":"Using Nifs - Yielding nifs","type":"extras"},{"doc":"If you want to launch your nif as threaded, use the threaded attribute. Note that this a heavy handed operation that is inappropriate if you can do yielding or dirty nifs. You should use this strategy if:you are using 3rd party library code that can't take advantage of yielding nifs.you may need to run more nifs concurrently than you have dirty schedulersyou don't mind the overhead of spinning up an OS thread.~Z&quot;&quot;&quot; /// nif: threaded/0 threaded fn threaded(env: beam.env) beam.term { std.time.sleep(10_000_000); return beam.make_ok(env); } &quot;&quot;&quot; test &quot;threaded&quot; do assert :ok == threaded() end","ref":"nifs.html#threaded-nifs","title":"Using Nifs - Threaded nifs","type":"extras"},{"doc":"If you want to launch your nif as a dirty, use the dirty_cpu or dirty_io attributes. Note that by default, your vm will only have a limited number of dirty schedulers available and launching the nif may fail if all of them are occupied. You should use this strategy if:you are using 3rd party library code that can't take advantage of yielding nifs.you don't mind that dirty_cpu jobs will be queued by the system if there the request rate exceeds availability.~Z&quot;&quot;&quot; /// nif: dirty_cpu/0 dirty_cpu fn dirty_cpu(env: beam.env) beam.term { std.time.sleep(10_000_000); return beam.make_ok(env); } &quot;&quot;&quot; test &quot;dirty_cpu&quot; do assert :ok == dirty_cpu() end","ref":"nifs.html#dirty-nifs","title":"Using Nifs - Dirty nifs","type":"extras"},{"doc":"Using ResourcesResources are C ABI objects which you declare should be managed by the BEAM reference-counted garbage collector. If you are passing data between function calls, it is best practice to pass them as a resource. Note that Zigler does not currently support passing references between multiple Zigler modules.Important NoteThe examples in this guide build upon each other and are tested in an integration test; so ~Z blocks present build upon code that is declared in previous ~Z blocks.","ref":"resources.html","title":"Using Resources","type":"extras"},{"doc":"A resource is declared by using the /// resource: &lt;name&gt; definition directive. The name should match a zig type declared in the next line.ExamplesIn the following example, the resource i64_res is bound to the builtin i64 type. This would allow you to store a mutable 64-bit integer and pass its pointer between function calls in the BEAM. NOTE: although it's mutable it is NOT threadsafe.~Z&quot;&quot;&quot; /// resource: i64_res definition const i64_res = i64; &quot;&quot;&quot;In the next example, resource struct_res is bound to a custom struct type. This is a more typical situation, where structured data, possibly requiring internal memory allocation, needs to be passed between function calls.~Z&quot;&quot;&quot; /// resource: struct_res definition const struct_res = struct { first_name: []u8, last_name: []u8, }; &quot;&quot;&quot;","ref":"resources.html#basics-definition","title":"Using Resources - Basics: Definition","type":"extras"},{"doc":"In order to create a resource, a __resource__ struct is autogenerated which makes manipulating BEAM resources much simpler. When returned from the nif, the resource will appear as a reference/0 term. As this term is passed around, potentially between processes, it will only be garbage collected when all references have been to the term have been lost (more on that in cleanup).The create function takes three parameters. First, the type that is being encapsulated, second, the BEAM environment of the calling function, and third, the initial value for those parameters.Important notesYou zig code will only have access to the __resource__ struct in the context of NIF code associated with your module in which it resides, within the Zig.sigil_z/2 blocks; you cannot use the __resource__ struct in other modules or zig code which has been @included.Erlang/Elixir code will have no direct visibility of the content of these resources: you MUST use nif functions to access or modify those contents.Resource references must remain local. Unlike other erlang terms, the content will not be automatically serialized, transferred to another node, and deserialized on the other side.Resource creation is failable; you must catch the resource error and handle it before returning the resource term. In a future release it will be possible to pass it upwards with the try function with a full error return trace.For datastructures (structs, slices) you have the choice of storing it directly as the resource, or storing the pointer to the datastructure as the resource.Example: Resource Creation~Z&quot;&quot;&quot; /// nif: create_i64/0 fn create_i64(env: beam.env) beam.term { return __resource__.create(i64_res, env, 47) catch return beam.raise_resource_error(env); } /// nif: create_struct/2 fn create_struct(env: beam.env, first_name: []u8, last_name: []u8) beam.term { var resource_fn = beam.allocator.alloc(u8, first_name.len) catch return beam.raise_enomem(env); errdefer beam.allocator.free(resource_fn); var resource_ln = beam.allocator.alloc(u8, last_name.len) catch return beam.raise_enomem(env); errdefer beam.allocator.free(resource_ln); std.mem.copy(u8, resource_fn, first_name); std.mem.copy(u8, resource_ln, last_name); return __resource__.create(struct_res, env, .{ .first_name = resource_fn, .last_name = resource_ln, }) catch return beam.raise_resource_error(env); } &quot;&quot;&quot; test &quot;resource creation generates references&quot; do i64_res_example = create_i64() assert is_reference(i64_res_example) struct_res_example = create_struct(&quot;john&quot;, &quot;doe&quot;) assert is_reference(struct_res_example) endExample: Resource Pointer Creation~Z&quot;&quot;&quot; /// resource: struct_ptr_res definition const struct_ptr_res = *struct_res; /// nif: create_struct_ptr/2 fn create_struct_ptr(env: beam.env, first_name: []u8, last_name: []u8) beam.term { var resource = beam.allocator.create(struct_res) catch return beam.raise_enomem(env); errdefer beam.allocator.destroy(resource); resource.first_name = beam.allocator.alloc(u8, first_name.len) catch return beam.raise_enomem(env); errdefer beam.allocator.free(resource_fn); resource.last_name = beam.allocator.alloc(u8, last_name.len) catch return beam.raise_enomem(env); errdefer beam.allocator.free(resource_ln); std.mem.copy(u8, resource.first_name, first_name); std.mem.copy(u8, resource.last_name, last_name); // note the signature vv here vv return __resource__.create(struct_ptr_res, env, resource) catch return beam.raise_resource_error(env); } &quot;&quot;&quot; test &quot;resource pointer generation creates a reference&quot; do struct_ptr_example = create_struct_ptr(&quot;john&quot;, &quot;doe&quot;) assert is_reference(struct_ptr_example) end","ref":"resources.html#creation","title":"Using Resources - Creation","type":"extras"},{"doc":"The __resource__ struct has fetch and update functions which allow you to retrieve or modify the contents in the struct, respectively. Note that although you can potentially access these data from different processes, it is not threadsafe by default without implementing mutexes.The fetch function takes three parameters, the internal type of the resource, the calling function environment, and the resource beam.term value.The update function takes four parameters, the internal type of the resource, the calling function environment, the resource beam.term value, and the replacement value.Important considerationsResource fetching is failable, and you should handle it by returning a beam raise term. This may not necessarily be the same type as the contents of the resource itself! So you might have to marshall it into a term manually. Note: this code will likely become simpler in the future.The data encapsulated by the Resource term are mutable! If you update the term, that change will persist within the same call scope, across function call scopes, etc! Also importantly, if you pass the data to another process you must take care to guard changes around mutexes and be aware that change or could be subject to unexpected race conditions without additional coordination.For structs, you must replace the entire contents of the struct; depending on your needs, you may want to consider using a pointer, which will let you change the contents of the struct WITHOUT performing an update. Note that extra care must be taken in this case to avoid race conditions resulting in use-after-free segfaults.Example: Fetching~Z&quot;&quot;&quot; /// nif: fetch_i64/1 fn fetch_i64(env: beam.env, res: beam.term) beam.term { var result = __resource__.fetch(i64_res, env, res) catch return beam.raise_resource_error(env); return beam.make_i64(env, result); } /// nif: fetch_struct_fn/1 fn fetch_struct_fn(env: beam.env, res: beam.term) beam.term { var result = __resource__.fetch(struct_res, env, res) catch return beam.raise_resource_error(env); return beam.make_slice(env, result.first_name); } /// nif: fetch_struct_ln/1 fn fetch_struct_ln(env: beam.env, res: beam.term) beam.term { var result = __resource__.fetch(struct_res, env, res) catch return beam.raise_resource_error(env); return beam.make_slice(env, result.last_name); } // note that the function bodies for the struct pointer form look NEARLY // identical to that of the direct struct resource form. /// nif: fetch_struct_ptr_fn/1 fn fetch_struct_ptr_fn(env: beam.env, res: beam.term) beam.term { var result = __resource__.fetch(struct_ptr_res, env, res) catch return beam.raise_resource_error(env); return beam.make_slice(env, result.first_name); } /// nif: fetch_struct_ptr_ln/1 fn fetch_struct_ptr_ln(env: beam.env, res: beam.term) beam.term { var result = __resource__.fetch(struct_ptr_res, env, res) catch return beam.raise_resource_error(env); return beam.make_slice(env, result.last_name); } &quot;&quot;&quot; test &quot;resources can be fetched&quot; do i64_res_example = create_i64() assert 47 == fetch_i64(i64_res_example) struct_res_example = create_struct(&quot;john&quot;, &quot;doe&quot;) assert &quot;john&quot; == fetch_struct_fn(struct_res_example) assert &quot;doe&quot; == fetch_struct_ln(struct_res_example) struct_ptr_res_example = create_struct_ptr(&quot;john&quot;, &quot;doe&quot;) assert &quot;john&quot; == fetch_struct_ptr_fn(struct_ptr_res_example) assert &quot;doe&quot; == fetch_struct_ptr_ln(struct_ptr_res_example) endExample: Updating~Z&quot;&quot;&quot; /// nif: update_i64/2 fn update_i64(env: beam.env, res: beam.term, update_value: i64) beam.term { __resource__.update(i64_res, env, res, update_value) catch return beam.raise_resource_error(env); return beam.make_ok(env); } /// nif: update_struct_fn/2 fn update_struct_fn(env: beam.env, res: beam.term, update_value: []u8) beam.term { var old_value = __resource__.fetch(struct_res, env, res) catch return beam.raise_resource_error(env); var new_first_name = beam.allocator.alloc(u8, update_value.len) catch return beam.raise_enomem(env); errdefer beam.allocator.free(new_first_name); std.mem.copy(u8, new_first_name, update_value); defer beam.allocator.free(old_value.first_name); var new_struct = .{ .first_name = new_first_name, .last_name = old_value.last_name, }; __resource__.update(struct_res, env, res, new_struct) catch return beam.raise_resource_error(env); return beam.make_ok(env); } /// nif: update_struct_ln/2 fn update_struct_ln(env: beam.env, res: beam.term, update_value: []u8) beam.term { var old_value = __resource__.fetch(struct_res, env, res) catch return beam.raise_resource_error(env); var new_last_name = beam.allocator.alloc(u8, update_value.len) catch return beam.raise_enomem(env); errdefer beam.allocator.free(new_last_name); std.mem.copy(u8, new_last_name, update_value); defer beam.allocator.free(old_value.last_name); var new_struct = .{ .first_name = old_value.first_name, .last_name = new_last_name, }; __resource__.update(struct_res, env, res, new_struct) catch return beam.raise_resource_error(env); return beam.make_ok(env); } // note that updating POINTER structs does not require calling update. /// nif: update_struct_ptr_fn/2 fn update_struct_ptr_fn(env: beam.env, res: beam.term, update_value: []u8) beam.term { var struct_ptr = __resource__.fetch(struct_ptr_res, env, res) catch return beam.raise_resource_error(env); beam.allocator.free(struct_ptr.first_name); struct_ptr.first_name = beam.allocator.alloc(u8, update_value.len) catch return beam.raise_enomem(env); errdefer beam.allocator.free(struct_ptr.first_name); std.mem.copy(u8, struct_ptr.first_name, update_value); return beam.make_ok(env); } /// nif: update_struct_ptr_ln/2 fn update_struct_ptr_ln(env: beam.env, res: beam.term, update_value: []u8) beam.term { var struct_ptr = __resource__.fetch(struct_ptr_res, env, res) catch return beam.raise_resource_error(env); beam.allocator.free(struct_ptr.last_name); struct_ptr.last_name = beam.allocator.alloc(u8, update_value.len) catch return beam.raise_enomem(env); errdefer beam.allocator.free(struct_ptr.last_name); std.mem.copy(u8, struct_ptr.last_name, update_value); return beam.make_ok(env); } &quot;&quot;&quot; test &quot;resources can be updated&quot; do i64_res_example = create_i64() assert :ok == update_i64(i64_res_example, 50) assert 50 == fetch_i64(i64_res_example) struct_res_example = create_struct(&quot;john&quot;, &quot;doe&quot;) assert :ok == update_struct_fn(struct_res_example, &quot;jane&quot;) assert &quot;jane&quot; == fetch_struct_fn(struct_res_example) assert &quot;doe&quot; == fetch_struct_ln(struct_res_example) assert :ok == update_struct_ln(struct_res_example, &quot;smith&quot;) assert &quot;jane&quot; == fetch_struct_fn(struct_res_example) assert &quot;smith&quot; == fetch_struct_ln(struct_res_example) struct_res_example = create_struct_ptr(&quot;john&quot;, &quot;doe&quot;) assert :ok == update_struct_ptr_fn(struct_res_example, &quot;jane&quot;) assert &quot;jane&quot; == fetch_struct_ptr_fn(struct_res_example) assert &quot;doe&quot; == fetch_struct_ptr_ln(struct_res_example) assert :ok == update_struct_ptr_ln(struct_res_example, &quot;smith&quot;) assert &quot;jane&quot; == fetch_struct_ptr_fn(struct_res_example) assert &quot;smith&quot; == fetch_struct_ptr_ln(struct_res_example) end","ref":"resources.html#fetching-and-updating","title":"Using Resources - Fetching and Updating","type":"extras"},{"doc":"If you are using resources in threaded or yielding long-running nifs, you should perform keep and release operations on these resources. This will prevent the garbage collector from triggering cleanup on your resource in the event that the calling process has been terminated. If you don't perform keep/release on the resource, it will at some point result in a segfault due to use-after-free and it may be difficult to detect because triggering segfault may depend on race conditions, VM computational load, and memory page layout.keep is failable and takes out a gc reference to your resource; it should always be matched by a release call (typically with a defer statement). Release is not failable.Both functions take three parameters, the type of the internal data structure, the calling function environment, and the resource term.Example: Keep and Release~Z&quot;&quot;&quot; /// nif: long_running_i64/1 threaded fn long_running_i64(env: beam.env, res: beam.term) beam.term { __resource__.keep(i64_res, env, res) catch return beam.raise_resource_error(env); defer __resource__.release(i64_res, env, res); var int_value = __resource__.fetch(i64_res, env, res) catch return beam.raise_resource_error(env); return beam.make_i64(env, int_value); } &quot;&quot;&quot; test &quot;keep and release&quot; do i64_res_example = create_i64() assert 47 == long_running_i64(i64_res_example) end","ref":"resources.html#keep-and-release","title":"Using Resources - Keep and Release","type":"extras"},{"doc":"Most resources should define a cleanup function. This function will be invoked when the last reference to that resource has been garbage collected by the virtual machine. The cleanup function is declared by using the /// resource: &lt;name&gt; cleanup, this should be followed by a void function taking a pointer to the resource. If there are any internal data that need freeing, you should do it here.Important NoteThe pointer for the resource itself will be cleaned up by Zigler, so don't free that.If your resource is a struct pointer, DO clean that up.Example: Cleanup~Z&quot;&quot;&quot; /// resource: struct_res cleanup fn struct_res_cleanup(env: beam.env, resource: *struct_res) void { beam.allocator.free(resource.first_name); beam.allocator.free(resource.last_name); // don&#39;t clean up the resource pointer itself. } /// resource: struct_res_ptr cleanup fn struct_res_ptr_cleanup(env: beam.env, resource: *struct_res_ptr) void { beam.allocator.free(resource.first_name); beam.allocator.free(resource.last_name); // DO clean up the pointer in the resource. beam.allocator.destroy(resource.*); // don&#39;t clean up the resource pointer (pointer). } &quot;&quot;&quot;","ref":"resources.html#cleanup","title":"Using Resources - Cleanup","type":"extras"}]